/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is ForumZilla 0.1.1.
 *
 * The Initial Developer of the Original Code is Zapogee, Inc.
 * Portions created by Zapogee, Inc. are Copyright (C) 2001
 * Zapogee, Inc.  All Rights Reserved.
 *
 * Contributor(s):
 *   Myk Melez, <myk@melez.com>, original author
 */


//****************************************************************************//
// constants and global variables

var gFeedService;

// The set of feeds currently being loaded, indexed by feed ID (resource URI).
// XXX Perhaps we should reuse the feed service's collection and just flag feeds
// in the process of being downloaded.
var loadCache = {};


//****************************************************************************//
// utility functions

function getNodeValue(node) {
  if (node && node.textContent)
    return node.textContent;
  else if (node && node.firstChild) {
    var ret = "";
    for (var child = node.firstChild; child; child = child.nextSibling) {
      var value = getNodeValue(child);
      if (value)
        ret += value;
    }

    if (ret)
      return ret;
  }

  return null;
}

function getRDFTargetValue(ds, source, property) {
  var node = ds.GetTarget(source, property, true);
  if (node) {
    node = node.QueryInterface(Components.interfaces.nsIRDFLiteral);
    if (node)
      return node.Value;
  }
  return null;
}


//****************************************************************************//
// nsFeed component

function Feed() {
    if (!gFeedService)
        gFeedService = 
          Components
            .classes["@forumzilla.mozdev.org/feed/feedservice;1"]
              .getService(Components.interfaces.nsIFeedService);
}

Feed.interfaces = [Components.interfaces.nsIFeed,
                   Components.interfaces.nsISupports];

Feed.prototype.QueryInterface = function (iid) {
    for ( var i=0 ; i<Feed.interfaces.length ; i++) {
        if (iid.equals(Feed.interfaces[i]))
            return this;
    }

    Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
    return null;
}


//****************************************************************************//
// nsIFeed interface

Feed.prototype.description = null;
Feed.prototype.author = null;
Feed.prototype.request = null;

Feed.prototype.init = function(resource) {
    //debug("Feed.init: resource=" + resource);
    this.resource = resource.QueryInterface(Components.interfaces.nsIRDFResource);
}

// XXX use this in feeditem as well
Feed.prototype.updateField = function(field, oldValue, newValue) {
    if (oldValue)
        this.ds.Change(this.resource, field, oldValue, newValue);
    else
        this.ds.Assert(this.resource, field, newValue, true);
}

Feed.prototype.__defineGetter__('url', function() {
    // We can't use this.debug in this getter or anything this getter calls
    // because this.debug calls this.url, so calling this.debug here would result
    // in infinite recursion.
    var url = this.ds.GetTarget(this.resource, DC_IDENTIFIER, true);
    if (url)
      url = url.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
    else
      url = this.resource.Value;
    return url;
});

Feed.prototype.__defineSetter__('url', function(newValue) {
    var oldValue = this.ds.GetTarget(this.resource, DC_IDENTIFIER, true);
    newValue = rdf.GetLiteral(newValue);
    this.updateField(DC_IDENTIFIER, oldValue, newValue);
    
    // Make sure feed service caches this feed at the new URL.
    // XXX The feed service should have a more direct API for this.
    gFeedService.getFeed(this.resource);

    // XXX Update feeditems data source with new URL for references to this feed.
});

Feed.prototype.__defineGetter__('ds', function() { return gFeedService.subscriptionsDS });


// XXX We shouldn't store a resource, we should just call GetResource
// on the URL instead.

Feed.prototype.__defineGetter__('resource', function() {
    return this._resource || rdf.GetResource(this.url);
});

Feed.prototype._resource = null;

Feed.prototype.__defineSetter__('resource', function(resource) {
    this._resource = resource;
});


Feed.prototype.__defineGetter__('title', function() {
    var ds = gFeedService.subscriptionsDS;
    var title = ds.GetTarget(this.resource, DC_TITLE, true);
    if (title)
      title = title.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
    return title;
});

Feed.prototype.__defineSetter__('title', function(newValue) {
    var ds = gFeedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, DC_TITLE, true);
    newValue = rdf.GetLiteral(newValue);
    this.updateField(DC_TITLE, oldValue, newValue);
});

Feed.prototype.__defineGetter__('quickMode', function() {
    var ds = gFeedService.subscriptionsDS;
    var quickMode = ds.GetTarget(this.resource, FZ_QUICKMODE, true);
    if (quickMode) {
        quickMode = quickMode.QueryInterface(Components.interfaces.nsIRDFLiteral);
        quickMode = quickMode.Value;
        quickMode = quickMode == "true" ? true : false;
    }
    return quickMode;
});

Feed.prototype.__defineSetter__('quickMode', function(newValue) {
    var ds = gFeedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, FZ_QUICKMODE, true);
    newValue = newValue ? "true" : "false";
    newValue = rdf.GetLiteral(newValue);
    this.updateField(FZ_QUICKMODE, oldValue, newValue);
});

Feed.prototype.__defineGetter__('enabled', function() {
    var ds = gFeedService.subscriptionsDS;
    var enabled = ds.GetTarget(this.resource, FZ_ENABLED, true);
    if (enabled) {
        enabled = enabled.QueryInterface(Components.interfaces.nsIRDFLiteral);
        enabled = enabled.Value;
        enabled = enabled == "true" ? true : false;
    }
    else
        enabled = true;
    return enabled;
});

Feed.prototype.__defineSetter__('enabled', function(newValue) {
    var ds = gFeedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, FZ_ENABLED, true);
    newValue = newValue ? "true" : "false";
    newValue = rdf.GetLiteral(newValue);
    this.updateField(FZ_ENABLED, oldValue, newValue);
});

// XXX The getter returns an nsIMsgFolder, but the setter takes a URI. Hmmm.
Feed.prototype.__defineGetter__('folder', function() {
    var ds = gFeedService.subscriptionsDS;
    var uri = ds.GetTarget(this.resource, FZ_FOLDER, true);

    if (uri) {
        uri = uri.QueryInterface(Components.interfaces.nsIRDFResource).Value;
        if (!gFeedService.mailWindow.GetMsgFolderFromUri(uri, true))
            gFeedService.createFolder(uri);
    }
    else {
        this.debug("no folder registered; checking for old school folder");
        uri = this.oldFolderURI;
        if (!uri) {
            this.debug("no old school folder; creating default folder");
            uri = gFeedService.createDefaultFolder(this.title);
        }
        this.debug("registering folder " + uri);
        // XXX This may fail if the RDF service thinks the URI is illegal;
        // figure out when this happens (Mozilla on Windows is my only clue)
        // and how to deal with it.
        ds.Assert(this.resource, FZ_FOLDER, rdf.GetResource(uri), true);
    }
    return gFeedService.mailWindow.GetMsgFolderFromUri(uri, true);
});

Feed.prototype.__defineSetter__('folder', function(newValue) {
    var ds = gFeedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, FZ_FOLDER, true);
    // XXX URI may not be Unicode char friendly, so we might have better luck
    // trying to QueryInterface the folder to nsIRDFResource, which it might
    // have an interface for.
    //newValue = rdf.GetResource(newValue.URI);
    newValue = newValue.QueryInterface(Components.interfaces.nsIRDFResource);
    this.updateField(FZ_FOLDER, oldValue, newValue);
});

// This returns a time double, but the setter receives a Date object; hmm.
Feed.prototype.__defineGetter__('lastPolled', function() {
    var ds = gFeedService.subscriptionsDS;
    var lastPolled = ds.GetTarget(this.resource, FZ_LASTPOLLED, true);
    if (lastPolled)
      lastPolled = new Date(lastPolled.QueryInterface(Components.interfaces.nsIRDFLiteral).Value).getTime();
    else
      lastPolled = 0;
    return lastPolled;
});

Feed.prototype.__defineSetter__('lastPolled', function(newValue) {
    var ds = gFeedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, FZ_LASTPOLLED, true);
    newValue = newValue.toUTCString();
    newValue = rdf.GetLiteral(newValue);
    this.updateField(FZ_LASTPOLLED, oldValue, newValue);
});

Feed.prototype.__defineGetter__('interval', function() {
    var ds = gFeedService.subscriptionsDS;
    var interval = ds.GetTarget(this.resource, FZ_INTERVAL, true);
    if (interval)
      interval = interval.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
    else
      interval = gFeedService.pollingInterval;
    return interval;
});

Feed.prototype.__defineSetter__('interval', function(newValue) {
    var ds = gFeedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, FZ_INTERVAL, true);
    newValue = rdf.GetLiteral(newValue);
    this.updateField(FZ_INTERVAL, oldValue, newValue);
});

// The name of the feed folder back when they had to be named after their feeds.
// XXX This can go away once enough users have upgraded from 0.5.0-.
Feed.prototype.__defineGetter__('oldFolderName', function() {
    var name = this.title || this.description || this.url;
    if (!name)
        throw("couldn't compute feed name, as feed has no title, description, or URL.");
  
    // Make sure the feed name doesn't have any line breaks, since we're going
    // to use it as the name of the folder in the filesystem.  This may not
    // be necessary, since Mozilla's mail code seems to handle other forbidden
    // characters in filenames and can probably handle these as well.
    name = name.replace(/[\n\r\t]+/g, " ");
  
    // Make sure the feed doesn't end in a period to work around bug 117840.
    name = name.replace(/\.+$/, "");
  
    return name;
});

// The URI of the feed folder back when they had to be named after their feeds.
// XXX This can go away once enough users have upgraded from 0.5.0-.
Feed.prototype.__defineGetter__('oldFolderURI', function() {
    var uri;
    try {
        // Try to get the folder by looking up the old pre-0.5.1 "News & Blogs"
        // account and seeing if there's a folder in there named after the feed.
        var accountManager =
            Components
                .classes["@mozilla.org/messenger/account-manager;1"]
                    .getService();
        accountManager = accountManager.QueryInterface(Components.interfaces.nsIMsgAccountManager);
        var accounts = accountManager.accounts.QueryInterface(Components.interfaces.nsICollection);
        var account, oldAccount;
        for ( var i=0 ; i<accounts.Count() ; i++ ) {
          account = accounts.GetElementAt(i);
          account = account.QueryInterface(Components.interfaces.nsIMsgAccount);
          if (account.incomingServer && account.incomingServer.prettyName == "News & Blogs")
            oldAccount = account;
            continue;
        }
        if (oldAccount
            && oldAccount.incomingServer
            && oldAccount.incomingServer.rootMsgFolder
            && oldAccount.incomingServer.rootMsgFolder.containsChildNamed(this.oldFolderName))
        {
            this.debug("old school folder for this feed; using it");
            uri = 
              oldAccount
                .incomingServer
                  .rootMsgFolder
                    .getChildNamed(this.oldFolderName)
                      .QueryInterface(Components.interfaces.nsIMsgFolder)
                        .URI;
        }
    }
    catch(e) {
        this.debug("error getting old school folder: " + e);
    }
    return uri;
});


Feed.prototype.__defineGetter__('charset', function() {
    if (!this.request || !this.request.responseXML)
        return null;
   return this.request.responseXML.characterSet;
});

// XXX We only ever use async and parseItems together when overloading
// this function to obtain the feed's title right after the user has subscribed
// to it.  Perhaps we should collapse these into a single parameter or have
// a separate getTitle or subscribe function to handle that.
Feed.prototype.download = function(async, parseItems) {
    // Whether or not to parse items when downloading and parsing the feed.
    // Setting to false is useful for obtaining just the title of the feed
    // when the user subscribes to it.
    this.parseItems = parseItems;
  
    // Mark the feed as polled so we don't try to download again
    // until the polling interval has passed.
    this.lastPolled = new Date();

    this.request =
      Components
        .classes["@mozilla.org/xmlextras/xmlhttprequest;1"]
          .createInstance(Components.interfaces.nsIJSXMLHttpRequest);
    this.request.open("GET", this.url, async);
    this.request.overrideMimeType("text/xml");
    if (async) {
        this.request.onload = Feed.onDownloaded;
        this.request.onerror = Feed.onDownloadError;
        // XXX Feeds should have an "ID" property that returns
        // the feed resource URI (this.resource.Value).
        loadCache[this.resource.Value] = this;
    }
    this.request.send(null);
    if (!async)
        this.parse();
}

Feed.onDownloaded = function(event) {
    var request = event.target;
    var url = request.channel.originalURI.spec;
    debug(url + " downloaded; retrieving feed object");

    // Retrieve the feed from the load cache.  We first try retrieving the feed
    // by URL, since feed IDs are usually the same as their URLs, then fall back
    // to searching the cache for a feed referencing the request.
    var feed = loadCache[url];
    if (!feed || feed.request != request) {
        for (var id in loadCache) {
            if (loadCache[id].request == request) {
                feed = loadCache[id];
                break;
            }
        }
    }
    if (!feed)
        throw(url + " error after downloading; couldn't retrieve feed");
    delete loadCache[feed.resource.Value];

    feed = feed.QueryInterface(Components.interfaces.nsIFeed);
    feed.parse();
}

Feed.onDownloadError = function(event) {
    var request = event.target;
    var url = request.channel.originalURI.spec;
    var status = "";
    var statusText = "";
    try {
        status = request.status;
        statusText = request.statusText;
    }
    catch(e) {}
    throw(url + " download failed: " + status + " " + statusText);
}

Feed.prototype.parse = function() {
    // Figures out what description language (RSS, Atom) and version this feed
    // is using and calls a language/version-specific feed parser.
  
    //var start = new Date();
    //this.debug("parsing feed " + this.url + " at " + start);
  
    if (!this.request.responseText) {
        throw("error parsing feed " + this.url + ": no data");
        return;
    }
    else if (this.request.responseText.search(/="http:\/\/purl\.org\/rss\/1\.0\/"/) != -1) {
        this.debug("is an RSS 1.x (RDF-based) feed");
        this.parseAsRSS1();
    }
    else if (this.request.responseXML.documentElement.namespaceURI == "http:\/\/purl.org\/atom\/ns#"
             || this.request.responseXML.documentElement.namespaceURI == "http://www.w3.org/2005/Atom"
             || this.request.responseXML.documentElement.nodeName == "feed"
             && !this.request.responseXML.documentElement.namespaceURI) {
        this.debug("is an Atom 0.3 or 1.0 feed");
        this.parseAsAtom();
    }
    else if (this.request.responseText.search(/"http:\/\/my\.netscape\.com\/rdf\/simple\/0\.9\/"/) != -1)
    {
        // RSS 0.9x is forward compatible with RSS 2.0, so use the RSS2 parser to handle it.
        this.debug("is an 0.9x feed");
        this.parseAsRSS2();
    }
    // XXX Explicitly check for RSS 2.0 instead of letting it be handled by the
    // default behavior (who knows, we may change the default at some point).
    else {
        // We don't know what kind of feed this is; let's pretend it's RSS 0.9x
        // and hope things work out for the best.  In theory even RSS 1.0 feeds
        // could be parsed by the 0.9x parser if the RSS namespace was the default.
        this.debug("is of unknown format; assuming an RSS 0.9x feed");
        this.parseAsRSS2();
    }
  
    //var end = new Date();
    //this.debug("done parsing feed " + this.url + " at " + end);
    //var cost = end - start;
    //gTotalCost += cost;
    //this.debug("cost for " + this.title + ": " + start + " to " + end + " = " + cost);
}

Feed.prototype.parseAsRSS2 = function() {
    if (!this.request.responseXML
        || !(this.request.responseXML instanceof Components.interfaces.nsIDOMXMLDocument))
      throw("error parsing RSS 2.0 feed " + this.url + ": data not parsed into XMLDocument object");

    // Get the first channel (assuming there is only one per RSS File).
    var channel = this.request.responseXML.getElementsByTagName("channel")[0];
    // If we can't find a channel element, look for stuff in the file as a whole.
    if (!channel)
        channel = this.request.responseXML;

    if (!this.title)
        this.title = getNodeValue(channel.getElementsByTagName("title")[0]);

    this.description = getNodeValue(channel.getElementsByTagName("description")[0]);

    if (!this.parseItems)
        return;

    this.invalidateItems();
  
    this.debug("parsing items");
    var itemNodes = this.request.responseXML.getElementsByTagName("item");
    for ( var i=0 ; i<itemNodes.length ; i++ ) {
        var itemNode = itemNodes[i];
        var item = 
          Components
            .classes["@forumzilla.mozdev.org/feed/feeditem;1"]
              .createInstance(Components.interfaces.nsIFeedItem);
        item.feed = this;
    
        var link = getNodeValue(itemNode.getElementsByTagName("link")[0]);
        //this.debug("link is " + link);
    
        var guidNode = itemNode.getElementsByTagName("guid")[0];
        var guid, isPermaLink;
        if (guidNode) {
            guid = getNodeValue(guidNode);
            isPermaLink =
              guidNode.getAttribute('isPermaLink') == 'false' ? false : true;
        }
    
        var url = (guid && isPermaLink) ? guid : link ? link : null;
        //this.debug("original url is " + url);
        if (link != url)
          this.debug("warning: original url != link");
        item.url = url;
        //this.debug("ultimate url is " + url);
        if (link != url)
          this.debug("warning: ultimate url != link");
        item.id = guid;
        item.description = getNodeValue(itemNode.getElementsByTagName("description")[0]);
        item.title = getNodeValue(itemNode.getElementsByTagName("title")[0])
                     || (item.description ? item.description.substr(0, 150) : null)
                     || item.title;
        item.author = getNodeValue(itemNode.getElementsByTagName("author")[0]
                                   || itemNode.getElementsByTagName("creator")[0]
                                   || itemNode.getElementsByTagName("source")[0])
                      || this.title
                      || item.author;
        item.date = getNodeValue(itemNode.getElementsByTagName("pubDate")[0]
                                 || itemNode.getElementsByTagName("date")[0]);
        var content = getNodeValue(itemNode.getElementsByTagNameNS(RSS_CONTENT_NS, "encoded")[0]);
        if (content)
            item.content = content;

        item.valid = true;
        if (!item.stored && !item.storing)
            item.store();
            //gFeedService.queueItem(item);
    }

    this.removeInvalidItems();
}

Feed.prototype.parseAsRSS1 = function() {
    // RSS 1.0 is valid RDF, so use the RDF parser/service to extract data.
  
    // Create a new RDF data source and parse the feed into it.
    var ds = Components
               .classes["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"]
                 .createInstance(Components.interfaces.nsIRDFDataSource);
    rdfparser.parseString(ds, this.request.channel.URI, this.request.responseText);
  
    // Get information about the feed as a whole.
    var channel = ds.GetSource(RDF_TYPE, RSS_CHANNEL, true);
    if (!this.title)
        this.title = getRDFTargetValue(ds, channel, RSS_TITLE);
    this.description = getRDFTargetValue(ds, channel, RSS_DESCRIPTION);

    if (!this.parseItems)
        return;
  
    this.invalidateItems();
  
    this.debug("parsing items");
    var items = ds.GetTarget(channel, RSS_ITEMS, true);
    try {
        if (!items)
            throw("no items container");
        //items = items.QueryInterface(Components.interfaces.nsIRDFContainer);
        items = rdfcontainer.MakeSeq(ds, items);
        items = items.GetElements();
        if (!items.hasMoreElements())
            throw("items container empty");
    }
    catch(e) {
        // For some reason we couldn't get a list of items from the items sequence,
        // (perhaps a buggy feed) so construct a list from all RSS item resources.
        items = ds.GetSources(RDF_TYPE, RSS_ITEM, true);
    }
  
    while (items.hasMoreElements()) {
        var itemResource = items.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
        var item = 
          Components
            .classes["@forumzilla.mozdev.org/feed/feeditem;1"]
              .createInstance(Components.interfaces.nsIFeedItem);
        item.feed = this;
    
        // Prefer the value of the link tag to the item URI since the URI could be
        // a relative URN.
        var uri = itemResource.Value;
        var link = getRDFTargetValue(ds, itemResource, RSS_LINK);
    
        item.url = link || uri;
        item.description = getRDFTargetValue(ds, itemResource, RSS_DESCRIPTION);
        item.title = getRDFTargetValue(ds, itemResource, RSS_TITLE)
                     || getRDFTargetValue(ds, itemResource, DC_SUBJECT)
                     || (item.description ? item.description.substr(0, 150) : null)
                     || item.title;
        item.author = getRDFTargetValue(ds, itemResource, DC_CREATOR)
                      || getRDFTargetValue(ds, channel, DC_CREATOR)
                      || this.title
                      || item.author;
        item.date = getRDFTargetValue(ds, itemResource, DC_DATE);
        var content = getRDFTargetValue(ds, itemResource, RSS_CONTENT_ENCODED);
        if (content)
            item.content = content;
        item.valid = true;
        if (!item.stored && !item.storing)
            item.store();
            //gFeedService.queueItem(item);
    }

    this.removeInvalidItems();
}

Feed.prototype.parseAsAtom = function() {
    if (!this.request.responseXML
        || !(this.request.responseXML instanceof Components.interfaces.nsIDOMXMLDocument))
      throw("error parsing Atom feed " + this.url + ": data not parsed into XMLDocument object");
  
    // Get the first channel (assuming there is only one per Atom File).
    var channel = this.request.responseXML.getElementsByTagName("feed")[0];
    if (!channel)
      throw("channel missing from Atom feed " + request.channel.name);

    if (!this.title)
        this.title = getNodeValue(channel.getElementsByTagName("title")[0]);
    this.description = getNodeValue(channel.getElementsByTagName("tagline")[0]);

    if (!this.parseItems)
      return;
  
    this.invalidateItems();
  
    this.debug("parsing items");
    var items = this.request.responseXML.getElementsByTagName("entry");
    for ( var i=0 ; i<items.length ; i++ ) {
        var itemNode = items[i];
        var item = 
          Components
            .classes["@forumzilla.mozdev.org/feed/feeditem;1"]
              .createInstance(Components.interfaces.nsIFeedItem);
        item.feed = this;
    
        var url;
        var links = itemNode.getElementsByTagName("link");
        for ( var j=0 ; j<links.length ; j++ ) {
            var alink = links[j];
            if (alink && alink.getAttribute('rel') && alink.getAttribute('rel') == 'alternate' && alink.getAttribute('href')) {
                url = alink.getAttribute('href');
                break;
            }
        }
    
        item.url = url;
        item.id = getNodeValue(itemNode.getElementsByTagName("id")[0]);
        item.description = getNodeValue(itemNode.getElementsByTagName("summary")[0]);
        item.title = getNodeValue(itemNode.getElementsByTagName("title")[0])
                     || (item.description ? item.description.substr(0, 150) : null)
                     || item.title;
    
        var author = itemNode.getElementsByTagName("author")[0]
                     || itemNode.getElementsByTagName("contributor")[0]
                     || channel.getElementsByTagName("author")[0];
        if (author) {
            var name = getNodeValue(author.getElementsByTagName("name")[0]);
            var email = getNodeValue(author.getElementsByTagName("email")[0]);
            if (name)
              author = name + (email ? " <" + email + ">" : "");
            else if (email)
              author = email;
            else
              author = null;
        }
        item.author = author || item.author || this.title;
    
        item.date = getNodeValue(itemNode.getElementsByTagName("modified")[0]
                                 || itemNode.getElementsByTagName("issued")[0]
                                 || itemNode.getElementsByTagName("created")[0]);
    
        // XXX We should get the xml:base attribute from the content tag as well
        // and use it as the base HREF of the message.
        // XXX Atom feeds can have multiple content elements; we should differentiate
        // between them and pick the best one.
        // Some Atom feeds wrap the content in a CTYPE declaration; others use
        // a namespace to identify the tags as HTML; and a few are buggy and put
        // HTML tags in without declaring their namespace so they look like Atom.
        // We deal with the first two but not the third.
        var content;
        var contentNode = itemNode.getElementsByTagName("content")[0];
        if (contentNode) {
            content = "";
            for ( j=0 ; j<contentNode.childNodes.length ; j++ ) {
                var node = contentNode.childNodes.item(j);
                if (node.nodeType == node.CDATA_SECTION_NODE)
                    content += node.data;
                else
                    content += serializer.serializeToString(node);
                    //content += getNodeValue(node);
            }
            if (contentNode.getAttribute('mode') == "escaped") {
                content = content.replace(/&lt;/g, "<");
                content = content.replace(/&gt;/g, ">");
                content = content.replace(/&amp;/g, "&");
            }
            if (content == "")
                content = null;
        }
        item.content = content;
    
        item.valid = true;
        if (!item.stored && !item.storing)
            item.store();
            //gFeedService.queueItem(item);
    }
  
    this.removeInvalidItems();
}

Feed.prototype.invalidateItems = function invalidateItems() {
    var ds = gFeedService.itemsDS;
    this.debug("invalidating items");
    var items = ds.GetSources(FZ_FEED, this.resource, true);
    while (items.hasMoreElements()) {
        var item = items.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
        this.debug(item.Value + " setting valid: false");
        var valid = ds.GetTarget(item, FZ_VALID, true);
        if (valid) {
            ds.Unassert(item, FZ_VALID, valid, true);
        }
        if (!ds.hasArcOut(item, FZ_INVALIDATED)) {
            var now = rdf.GetLiteral(new Date().toUTCString());
            ds.Assert(item, FZ_INVALIDATED, now, true);
            this.debug("setting invalidated: " + now.Value);
        }
    }
}

Feed.prototype.removeInvalidItems = function() {
    var ds = gFeedService.itemsDS;
    this.debug("removing invalid items");
    var items = ds.GetSources(FZ_FEED, this.resource, true);
    var item;
    while (items.hasMoreElements()) {
        item = items.getNext();
        item = item.QueryInterface(Components.interfaces.nsIRDFResource);

        if (ds.HasAssertion(item, FZ_VALID, RDF_LITERAL_TRUE, true)) {
            this.debug(item.Value + " not removing; valid");
            continue;
        }

        var invalidated = ds.GetTarget(item, FZ_INVALIDATED, true);
        if (invalidated) {
            var date = new Date(invalidated.QueryInterface(Components.interfaces.nsIRDFLiteral).Value);
            var time = date.getTime();
            if (new Date().getTime() - time < (1000*60*60*24)) {
                this.debug(item.Value + " not removing; too soon; invalidated " + date.toUTCString());
                continue;
            }
        }

        this.debug(item.Value + " removing");
        ds.Unassert(item, FZ_FEED, this.resource, true);
        if (ds.hasArcOut(item, FZ_FEED))
            this.debug(item.Value + " from more than one feed; removing this one");
        else
            removeAssertions(ds, item);
    }
}

Feed.prototype.debug = function(txt) {
    debug(this.url + " " + txt);
}


//****************************************************************************//
// module registration; adapted from nsUpdateService.js

var gModule = {
    _firstTime: true,
    _scriptsLoaded: false,
    
    _objects: {
        primary: {
            CID: Components.ID("{713ccd0c-ca01-4601-8970-255479317902}"),
            contractID: "@forumzilla.mozdev.org/feed/feed;1",
            className: "nsFeed",
            factory: {
                createInstance: function (aOuter, aIID) {
                    if (aOuter != null)
                        throw Components.results.NS_ERROR_NO_AGGREGATION;
                    return new Feed().QueryInterface(aIID);
                }
            }
        }
    },

    _loadScripts: function () {
        if (this._scriptsLoaded)
            return;

        var scriptLoader =
            Components
              .classes["@mozilla.org/moz/jssubscript-loader;1"]
                .getService(Components.interfaces.mozIJSSubScriptLoader);

        if (!scriptLoader)
            return;

        // Gives us LocalFile, an easy interface to files on the file system.
        scriptLoader.loadSubScript("chrome://forumzilla/content/file-utils.js");
        scriptLoader.loadSubScript("chrome://forumzilla/content/global.js");
        scriptLoader.loadSubScript("chrome://forumzilla/content/debug-utils.js");

        this._scriptsLoaded = true;

        gFeedService = 
          Components
            .classes["@forumzilla.mozdev.org/feed/feedservice;1"]
              .getService(Components.interfaces.nsIFeedService);

    },

    registerSelf: function (aComponentManager, aFileSpec, aLocation, aType) {
        if (this._firstTime) {
            this._firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }
        aComponentManager =
            aComponentManager
                .QueryInterface(Components.interfaces.nsIComponentRegistrar);
        
        for (var key in this._objects) {
            var obj = this._objects[key];
            aComponentManager.registerFactoryLocation(obj.CID,
                                                      obj.className,
                                                      obj.contractID,
                                                      aFileSpec,
                                                      aLocation,
                                                      aType);
        }
    },
    
    getClassObject: function (aComponentManager, aCID, aIID) {
        if (!aIID.equals(Components.interfaces.nsIFactory))
            throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
        
        if (!this._scriptsLoaded)
            this._loadScripts();
        
        for (var key in this._objects) {
            if (aCID.equals(this._objects[key].CID))
              return this._objects[key].factory;
        }
        
        throw Components.results.NS_ERROR_NO_INTERFACE;
    },
  
    canUnload: function (aComponentManager) {
        return true;
    }
};

function NSGetModule(compMgr, fileSpec) {
    return gModule;
}
